unit ExprProperty;

// Property editors for integers and floating point numbers
// where the user can enter a simple expression instead of just
// a number. The expression follows the following rules, which
// are roughly the same as those of Pascal:
//
// Expr ::= term + term | term - term | term OR term | term XOR term
// Term ::= factor * factor | factor / factor |
//          factor DIV factor | factor MOD factor |
//          factor AND factor | factor SHR factor | factor SHL factor
// Factor ::= ( Expr ) | - expr | + expr | NOT expr | primary
// Primary ::= number | identifier
//
// An identifier is the name of another property of the same object.
// Numbers are floating point internally, and are converted to
// integers when needed.
//
// Copyright  1999 Tempest Software, Inc.

interface

uses Windows, SysUtils, Classes, DsgnIntf;

type
  TIntegerExpressionProperty = class(TIntegerProperty)
  public
    procedure SetValue(const Str: string); override;
  end;
  TFloatExpressionProperty = class(TFloatProperty)
  public
    procedure SetValue(const Str: string); override;
  end;

  TParseTokenKind = (ptNone, ptEOF,
                     ptAdd, ptSub, ptMul, ptFDiv, ptDiv, ptMod,
                     ptAnd, ptOr, ptXor, ptNot, ptShr, ptShl,
                     ptLParen, ptRParen,
                     ptNumber, ptIdentifier);
  TParseToken = record
    Kind: TParseTokenKind;
    Text: string;
  end;
  TParseStream = class(TStream)
  private
    fStream: TStream;
    fPropOwner: TPersistent;
    fValue: Extended;
  protected
    Token: TParseToken;
    PushToken: TParseToken;
    procedure BackupStream;
    procedure BackupToken;
    function IsEof: Boolean;
    function SkipWhiteSpace: Boolean;
    function GetToken: Boolean;
    function GetExpr(var Value: Extended): Boolean;
    function GetTerm(var Value: Extended): Boolean;
    function GetFactor(var Value: Extended): Boolean;
    function GetProperty(const PropName: string): Extended;

    function GetTokenValue: Extended;

    property PropOwner: TPersistent read fPropOwner;
    property Stream: TStream read fStream;
  public
    constructor Create(PropOwner: TPersistent; Stream: TStream);
    function Parse: Boolean;
    function IntegerValue: Integer;
    function Int64Value: Int64;

    function Read(var Buffer; Count: LongInt): LongInt; override;
    function Write(const Buffer; Count: LongInt): LongInt; override;
    function Seek(Offset: LongInt; Origin: Word): LongInt; override;

    property Value: Extended read fValue;
  end;

procedure Register;

implementation

uses TypInfo;


{ TParseStream }

constructor TParseStream.Create(PropOwner: TPersistent; Stream: TStream);
begin
  inherited Create;
  fPropOwner := PropOwner;
  fStream := Stream;
end;

// Set the stream to reread the most recently read character.
procedure TParseStream.BackupStream;
begin
  Stream.Seek(-1, soFromCurrent);
end;

procedure TParseStream.BackupToken;
resourcestring
  sPushbackMultiple = 'Internal error in TParseStream.BackupToken: attempt to push back multiple tokens';
begin
  Assert(PushToken.Kind = ptNone, sPushbackMultiple);
  PushToken := Token;
end;

// Get an additive expression.
function TParseStream.GetExpr(var Value: Extended): Boolean;
resourcestring
  sMissingOperand = 'Syntax error: missing operand in expression';
var
  Operand: Extended;
  Operator: TParseTokenKind;
begin
  if not GetTerm(Value) then
    Result := False
  else
  begin
    Result := True;
    while GetToken do
    begin
      Operator := Token.Kind;
      if not (Operator in [ptAdd, ptSub, ptOr, ptXor]) then
      begin
        BackupToken;
        Exit;
      end;

      if not GetTerm(Operand) then
        raise Exception.Create(sMissingOperand);
      case Operator of
      ptAdd:
        Value := Value + Operand;
      ptSub:
        Value := Value - Operand;
      ptOr:
        Value := Round(Value) or Round(Operand);
      ptXor:
        Value := Round(Value) xor Round(Operand);
      end;
    end;
  end;
end;

function TParseStream.GetFactor(var Value: Extended): Boolean;
resourcestring
  sInvalidType = 'Internal error: invalid token type (%d) in GetFactor';
  sMissingRParen = 'Syntax error: unbalanced parentheses';
begin
  GetToken;
  case Token.Kind of
  ptEOF:
    Result := False;
  ptLParen:
    begin
      Result := GetExpr(Value) and GetToken;
      if not Result or (Token.Kind <> ptRParen) then
        raise Exception.Create(sMissingRParen);
    end;
  ptSub:
    begin
      Result := GetExpr(Value);
      if Result then
        Value := -Value;
    end;
  ptAdd:
    Result := GetExpr(Value);
  ptNot:
    begin
      Result := GetExpr(Value);
      if Result then
        Value := not Round(Value);
    end;
  ptNumber, ptIdentifier:
    begin
      Result := True;
      Value := GetTokenValue;
    end;
  else
    begin
      BackupToken;
      Result := True;
    end;
  end;
end;

function TParseStream.GetTerm(var Value: Extended): Boolean;
resourcestring
  sMissingTerm = 'Syntax error: missing term at end of expression';
var
  Operand: Extended;
  Operator: TParseTokenKind;
begin
  if not GetFactor(Value) then
    Result := False
  else
  begin
    Result := True;
    while GetToken do
    begin
      Operator := Token.Kind;
      if not (Operator in [ptMul, ptDiv, ptFDiv, ptMod, ptAnd, ptShr, ptShl]) then
      begin
        BackupToken;
        Break;
      end;

      if not GetFactor(Operand) then
        raise Exception.Create(sMissingTerm);
      case Operator of
      ptMul:
        Value := Value * Operand;
      ptFDiv:
        Value := Value / Operand;
      ptDiv:
        Value := Round(Value) div Round(Operand);
      ptMod:
        Value := Round(Value) mod Round(Operand);
      ptAnd:
        Value := Round(Value) and Round(Operand);
      ptShr:
        Value := Round(Value) shr Round(Operand);
      ptShl:
        Value := Round(Value) shl Round(Operand);
      end;
    end;
  end;
end;

function TParseStream.GetToken: Boolean;
resourcestring
  sUnknownOp = 'Unknown character: ''%s''';
const
  Alphabetic = ['a'..'z', 'A'..'Z', '_'];
  Digit = ['0'..'9'];
  HexDigit = Digit + ['a'..'f', 'A'..'F'];
  Alphanumeric = Alphabetic + Digit;
var
  C: Char;
begin
  if PushToken.Kind <> ptNone then
  begin
    Token := PushToken;
    PushToken.Kind := ptNone;
    Result := Token.Kind <> ptEOF;
    Exit;
  end;

  SkipWhiteSpace;
  if Stream.Read(C, 1) <> 1 then
  begin
    Token.Kind := ptEOF;
    Result := False;
  end
  else if C in Alphabetic then
  begin
    // Get the identifier name.
    Result := True;
    Token.Kind := ptIdentifier;
    Token.Text := C;
    while Stream.Read(C, 1) = 1 do
    begin
      if not (C in AlphaNumeric) then
      begin
        BackupStream;
        Break;
      end;
      Token.Text := Token.Text + C;
    end;

    // Special case tests for operator names.
    if CompareText(Token.Text, 'div') = 0 then
      Token.Kind := ptDiv
    else if CompareText(Token.Text, 'mod') = 0 then
      Token.Kind := ptMod
    else if CompareText(Token.Text, 'and') = 0 then
      Token.Kind := ptAnd
    else if CompareText(Token.Text, 'or') = 0 then
      Token.Kind := ptOr
    else if CompareText(Token.Text, 'not') = 0 then
      Token.Kind := ptNot
    else if CompareText(Token.Text, 'xor') = 0 then
      Token.Kind := ptXor
    else if CompareText(Token.Text, 'shr') = 0 then
      Token.Kind := ptShr
    else if CompareText(Token.Text, 'shl') = 0 then
      Token.Kind := ptShl
  end
  else if C = '$' then
  begin
    // Get the hexadecimal number.
    Result := True;
    Token.Kind := ptNumber;
    Token.Text := C;
    while Stream.Read(C, 1) = 1 do
    begin
      if not (C in HexDigit) then
      begin
        BackupStream;
        Break;
      end;
      Token.Text := Token.Text + C;
    end;
  end
  else if C in Digit then
  begin
    // Get the decimal number.
    Result := True;
    Token.Kind := ptNumber;
    Token.Text := C;
    while Stream.Read(C, 1) = 1 do
    begin
      if not (C in Digit) then
      begin
        BackupStream;
        Break;
      end;
      Token.Text := Token.Text + C;
    end;

    if not IsEof and (C = '.') then
    begin
      Stream.Read(C, 1);              // get the '.'
      Token.Text := Token.Text + C;

      // Get the fractional part of a floating point number.
      while Stream.Read(C, 1) = 1 do
      begin
        if not (C in Digit) then
        begin
          BackupStream;
          Break;
        end;
        Token.Text := Token.Text + C;
      end;
    end;

    if not IsEof and (C in ['e', 'E']) then
    begin
      Stream.Read(C, 1);              // get the 'e' or 'E'
      Token.Text := Token.Text + C;

      // Get the exponent part of a floating point number.
      if Stream.Read(C, 1) <> 1 then
        Exit;
      if C in ['+', '-'] then
        // Get the exponent sign;
        Token.Text := Token.Text + C
      else
        BackupStream;

      // Get the actual exponent.
      while Stream.Read(C, 1) = 1 do
      begin
        if not (C in Digit) then
        begin
          BackupStream;
          Break;
        end;
        Token.Text := Token.Text + C;
      end;
    end;
  end
  else
  begin
    // The token must be a single-character operator.
    Token.Text := C;
    Result := True;
    case C of
    '+': Token.Kind := ptAdd;
    '-': Token.Kind := ptSub;
    '*': Token.Kind := ptMul;
    '/': Token.Kind := ptFDiv;
    '(': Token.Kind := ptLParen;
    ')': Token.Kind := ptRParen;
    else
      raise Exception.CreateFmt(sUnknownOp, [C]);
    end; // outer case
  end
end;

function TParseStream.IsEof: Boolean;
begin
  Result := Stream.Position = Stream.Size;
end;

function TParseStream.Parse: Boolean;
begin
  Result := GetExpr(fValue);
  if GetToken then
    Result := False;  
end;

// Skip over white space and control characters.
// Return True if a non-white space character is available
// to be read, or False for end of file.
function TParseStream.SkipWhiteSpace: Boolean;
var
  C: Char;
begin
  while Stream.Read(C, 1) = 1 do
    if not (C in [#0..' ']) then
    begin
      BackupStream;
      Result := True;
      Exit;
    end;
  Result := False;
end;

// Look up an identifier as a property value.
// Raise an exception if the component does not have a property
// of that name, or if the property value cannot be converted to a number.
function TParseStream.GetProperty(const PropName: string): Extended;
resourcestring
  sNoProperty = 'No property named ''%s''';
  sPropNotNumber = 'Property ''%s'' is not numeric';
  sProperty = ' in property ''%s''';
var
  PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(PropOwner.ClassInfo, PropName);
  if PropInfo = nil then
    raise Exception.CreateFmt(sNoProperty, [PropName]);
  case PropInfo.PropType^.Kind of
  tkInteger, tkChar, tkWChar:
    Result := GetOrdProp(PropOwner, PropInfo);
  tkFloat:
    Result := GetFloatProp(PropOwner, PropInfo);
  tkString, tkLString, tkWString:
    try
      Result := StrToFloat(GetStrProp(PropOwner, PropInfo))
    except
    on Ex: EConvertError do
      raise EConvertError.Create(Ex.Message + Format(sProperty, [PropName]));
    else
      raise;
    end;
  else
    raise Exception.CreateFmt(sPropNotNumber, [PropName]);
  end;
end;

function TParseStream.GetTokenValue: Extended;
resourcestring
  sUnknownType = 'Internal error in TParseStream.GetTokenValue: Unknown token type (%d)';
begin
  case Token.Kind of
  ptIdentifier:
    Result := GetProperty(Token.Text);
  ptNumber:
    if Token.Text[1] = '$' then
      Result := StrToInt64(Token.Text)
    else
      Result := StrToFloat(Token.Text);
  else
    raise Exception.CreateFmt(sUnknownType, [Ord(Token.Kind)]);
  end;
end;

function TParseStream.Read(var Buffer; Count: Integer): LongInt;
begin
  Result := Stream.Read(Buffer, Count);
end;

function TParseStream.Seek(Offset: Integer; Origin: Word): LongInt;
begin
  Result := Stream.Seek(Offset, Origin);
end;

function TParseStream.Write(const Buffer; Count: Integer): LongInt;
begin
  Result := Stream.Write(Buffer, Count);
end;

function TParseStream.IntegerValue: Integer;
begin
  Result := Round(Value);
end;

function TParseStream.Int64Value: Int64;
begin
  Result := Round(Value);
end;



resourcestring
  sSyntaxError = 'Syntax error in expression';
  sRangeError = 'Expression result %d must be in the range [%d, %d]';

procedure RangeError(Value, MinValue, MaxValue: LongInt);
begin
  raise Exception.CreateFmt(sRangeError, [Value, MinValue, MaxValue]);
end;

{ TIntegerExpressionProperty }

procedure TIntegerExpressionProperty.SetValue(const Str: string);
var
  StringStream: TStringStream;
  ParseStream: TParseStream;
  Value: LongInt;
begin
  ParseStream := nil;
  StringStream := nil;
  try
    StringStream := TStringStream.Create(Str);
    ParseStream := TParseStream.Create(GetComponent(0), StringStream);
    if not ParseStream.Parse then
      raise Exception.Create(sSyntaxError);

    Value := ParseStream.IntegerValue;
    // Check that the value fits in the range of the property type.
    with GetTypeData(GetPropType)^ do
      if (Value < MinValue) or (Value > MaxValue) then
        RangeError(Value, MinValue, MaxValue);

{$R-}
    SetOrdValue(Value);
{$R+}
  finally
    ParseStream.Free;
    StringStream.Free;
  end;
end;

{ TFloatExpressionProperty }

procedure TFloatExpressionProperty.SetValue(const Str: string);
var
  StringStream: TStringStream;
  ParseStream: TParseStream;
begin
  ParseStream := nil;
  StringStream := nil;
  try
    StringStream := TStringStream.Create(Str);
    ParseStream := TParseStream.Create(GetComponent(0), StringStream);
    if not ParseStream.Parse then
      raise Exception.Create(sSyntaxError);
    SetFloatValue(ParseStream.Value);
  finally
    ParseStream.Free;
    StringStream.Free;
  end;
end;


procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(Integer), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(LongInt), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(Byte), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(Word), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(LongWord), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(SmallInt), TPersistent, '', TIntegerExpressionProperty);
  RegisterPropertyEditor(TypeInfo(ShortInt), TPersistent, '', TIntegerExpressionProperty);

  RegisterPropertyEditor(TypeInfo(Single), TPersistent, '', TFloatExpressionProperty);
  RegisterPropertyEditor(TypeInfo(Double), TPersistent, '', TFloatExpressionProperty);
  RegisterPropertyEditor(TypeInfo(Extended), TPersistent, '', TFloatExpressionProperty);
end;

end.
